#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import random
import os
import sys
import tty
import subprocess

from tvm.helpers import basehelper
from tvm.shared_helpers.tvstatistics import TVStatistics
from tvm.shared_helpers.io import err, info, info2
from tvm.shared_helpers.datehelpers import today_as_str, pretty_date


class PlayHelper(basehelper.BaseHelper):
    def __init__(self, config, collection):
        super(self.__class__, self).__init__(config, collection)

    def _increment_episode(self, episode):
        incremented_episode = str(int(episode[1].split(';')[-1]) + 1)
        incremented_episode_d = incremented_episode + ';' + str(int(incremented_episode) + 1)

        episode1, episode2 = list(episode), list(episode)
        episode1[1], episode2[1] = incremented_episode, incremented_episode_d

        return map(tuple, [episode1, episode2])

    def _increment_season(self, episode, reset_episode=False):
        """ `reset_episode=True` will set episode to 1. """
        episode = list(episode)
        episode[0] = str(int(episode[0]) + 1)

        if reset_episode:
            episode[1] = '1'

        return tuple(episode)

    def _find_next_episode(self, most_recent):
        # Current episode is '5', next could be '6' or '6;7'.
        next_ep1, next_ep2 = self._increment_episode(most_recent)

        if next_ep1 in self.collection.entries: return next_ep1
        if next_ep2 in self.collection.entries: return next_ep2

        # Finished the current season but not the entire show.
        next_season = self._increment_season(most_recent, reset_episode=True)
        next_ep1, next_pe2 = self._increment_episode(next_season)

        if next_season in self.collection.entries: return next_season

        # We are at the last episode and there are none left.
        return None

    def _build_episode_path(self, show):
        """ None is returned if unsuccessful. """
        season, episode, title = show
        for path in self.config.physical_storage_paths:
            possible_path = os.path.join(path, '/'.join([title, season]))

            # Get the file extension.
            for root, _, files in os.walk(possible_path):
                for file in files:
                    episode_number = os.path.splitext(file)[0]
                    if episode_number != episode:
                        continue

                    return os.path.join(possible_path, file)
        return None

    def update_records_with(self, episode):
        info('Recording "%s" as viewed.' % ','.join(episode))
        with open(self.config.viewed_path, 'a') as f:
            f.write(today_as_str() + ',%s,%s,"%s"\n' % episode)

    def play_episode(self, episode):
        """ Returns True if the episode was successfully played. """
        episode_path = self._build_episode_path(episode)
        if not episode_path:
            return False

        try:
            if subprocess.call(['open', '-a', 'VLC', episode_path]) == 0:
                self.update_records_with(episode)
                return True
        except OSError:
            err('Cannot open path because your platform "%s" is not supported.' % sys.platform)

        return False

    def next_to_watch(self, title):
        """ The `title` must be an exact title. @see find_matching_titles to find an exact match. """
        all_episodes, viewed_episodes = self.collection.find_all_episodes(title)

        # We have never seen any episodes yet.
        if not viewed_episodes:
            return self.collection.sort_episodes(all_episodes)[-1]

        # Sort by season, episode (descending order).
        viewed_episode = self.collection.sort_episodes(viewed_episodes)[0]

        # TODO: There is a bug here that causes this find_next to break if an episode is missing.
        #       find_next_episode definitely needs improvement.
        return self._find_next_episode(viewed_episode)

    def undo_last_played(self):
        if not len(self.collection.all_viewed_episodes):  # Nothing has yet to be viewed.
            return None

        # NOTE: Will not work for Windows.
        proc = subprocess.Popen(['tail', '-n-1', self.config.viewed_path], stdout=subprocess.PIPE)
        removed_episode = proc.stdout.read().strip('\n')

        proc = subprocess.Popen(['head', '-n-1', self.config.viewed_path], stdout=subprocess.PIPE)
        updated_file = proc.stdout.read()

        with open(self.config.viewed_path, 'w') as out_file:
            out_file.write(updated_file)

        return removed_episode

    def print_play_statistics(self, title):
        statistics = TVStatistics(self.collection)

        total_episodes_recorded = len(self.collection.entries)
        total_viewed_episodes = len(self.collection.all_viewed_episodes)
        total_percentage = float(total_viewed_episodes) / total_episodes_recorded * 100

        episodes_recorded, episodes_viewed = map(len, self.collection.find_all_episodes(title))
        show_percentage = float(episodes_viewed) / episodes_recorded * 100

        print  # New line to make it look less clumped together.
        info('Displaying basic play statistics:')
        info('%.2f%% (%d/%d), %.2f%% (%d/%d) views.' % (show_percentage, episodes_viewed, episodes_recorded,
                                                        total_percentage, total_viewed_episodes, total_episodes_recorded))

        first_recorded, _ = statistics.first_recorded_episode(title)
        last_updated, _ = statistics.last_updated_episode(title)
        last_viewed, _ = statistics.last_viewed_episode(title)

        info2('recorded_date: %s (%s).' % (pretty_date(first_recorded), first_recorded))
        info2('(updated: %s), (viewed: %s).' % (pretty_date(last_updated), pretty_date(last_viewed)))

    def _create_selection_pool(self, title):
        if title:
            # Return the all episodes as the selection pool if we're up to date with the show.
            recorded, viewed = self.collection.find_all_episodes(title)
            if self.collection.viewed_up_to_date(title):
                return recorded
            return viewed

        # A random selection on all recorded shows that have been viewed.
        return [episode[0] for episode in self.collection.all_viewed_episodes]

    def _get_random_episode(self, selection_pool, previously_randomised):
        if len(selection_pool) == len(previously_randomised):
            return False

        while True:
            episode = selection_pool[random.randint(0, len(selection_pool) - 1)]
            if episode not in previously_randomised:
                previously_randomised.append(episode)
                return episode

    def simulate_random_selection(self, title):
        selection_pool = self._create_selection_pool(title)
        previously_randomised = []

        info('Starting random selection simulation.')

        episode = self._get_random_episode(selection_pool, previously_randomised)
        tty.setcbreak(sys.stdin)
        while True:
            if not episode:
                return info('There are no episodes to randomly select from. Exiting.')

            info2('Play episode "%s" [y/n/q]? ' % ','.join(episode), end='')
            response = sys.stdin.read(1).lower()
            print(response.strip('\n'))  # Display what the user typed after the query.

            if response == 'y':
                info('Opening the episode, "%s".' % ','.join(episode))
                if not self.play_episode(episode):
                    err('Could not open episode "%s".' % ','.join(episode))
                    self.update_records_with(episode)
                return

            elif response == 'n':
                episode = self._get_random_episode(selection_pool, previously_randomised)
            elif response == 'q':
                return info('Leaving random selection simulation.')
