#!/usr/bin/env python
# -*- coding: utf-8 -*-
from config import Config

import csv
import os

from collections import defaultdict
from difflib import SequenceMatcher

from shared_helpers.datehelpers import convert_date
from shared_helpers.io import err, warn, info, ok
from shared_helpers.io import print_errors


class TVCollection(object):
    def __init__(self, config=None, ignore_errors=False, terminate_program=True):
        self._EXPECTED_RECORDS = ['shows_added.csv', 'shows_viewed.csv']
        self._ignore_errors = ignore_errors
        self._terminate_program = terminate_program

        self.config = Config() if config is None else config
        self._validate_and_fix_record_storage()
        self._load_records()

    def _record_path(self, record):
        return os.path.join(self.config.record_storage_path, record)

    def _parse_record(self, record):
        return tuple(record[1:]), convert_date(record[0])

    def _validate_and_fix_record_storage(self):
        """ Returns False if records don't exist, True otherwise. """
        shows_added_exists = os.path.exists(self._record_path('shows_added.csv'))
        shows_viewed_exists = os.path.exists(self._record_path('shows_viewed.csv'))

        # Critical error: Shows not added but shows are viewed.
        if not shows_added_exists and shows_viewed_exists:
            err('shows_viewed.csv was found but not shows_added.csv')
            exit(0)

        for record in self._EXPECTED_RECORDS:
            full_record_path = self._record_path(record)
            if not os.path.isfile(full_record_path):
                warn('"%s" does not exist. Creating for you.' % full_record_path)
                open(full_record_path, 'a').close()

    def _load_records(self):
        recorded_shows, errors = {}, defaultdict(list)
        with open(self._record_path('shows_added.csv'), 'rU') as recorded:
            for record in csv.reader(recorded):
                s_ep_title, date_recorded = self._parse_record(record)

                if s_ep_title in recorded_shows:  # Duplicate error.
                    errors['duplicates'].append(record)

                    # Ignoring errors will add the later most defined duplicate.
                    if not self._ignore_errors: continue

                # (season, episode, title) : ([dates_viewed, ], date_recorded)
                recorded_shows[s_ep_title] = ([], date_recorded)

        with open(self._record_path('shows_viewed.csv'), 'rU') as viewed:
            for record in csv.reader(viewed):
                s_ep_title, date_viewed = self._parse_record(record)

                if s_ep_title not in recorded_shows:
                    errors['viewed_not_recorded'].append(record)
                    if not self._ignore_errors: continue

                    # Adding it now because it wasn't recorded when shows_added.csv
                    if s_ep_title not in recorded_shows:
                        # NOTE: Let's assume that the date viewed was also when it was recorded.
                        #
                        #       This may cause problems with statistics in the future but if you want to
                        #       avoid inaccuracies... then you shouldn't be ignoring errors.
                        recorded_shows[s_ep_title] = ([date_viewed], date_viewed)
                        continue

                # There is always a date for when it was recorded.
                _, date_recorded = recorded_shows[s_ep_title]
                if date_viewed and date_viewed < date_recorded:
                    errors['viewed_before_recorded'].append(record)
                    if not self._ignore_errors: continue

                recorded_shows[s_ep_title][0].append(date_viewed)

        if self._ignore_errors and errors:
            warn('Errors found when loading recorded files. Continue with caution...')

        if not self._ignore_errors:
            print_errors(errors, terminate_program=self._terminate_program)

        self._load_success = False if errors else True
        self._recorded_shows = recorded_shows

    def find_matching_titles(self, query):
        """ `query` must be a string with words delimited by an underscore. """
        if not query:
            return None

        # Get the ratio of titles compared with the name.
        title_ratios = []
        for title in self.titles:
            ratio = round(SequenceMatcher(None, title, query).ratio(), 2)
            if ratio > 0.6 or (query in title and len(query) >= 3):
                # title, ratio, percentage
                title_ratios.append((title, ratio, int(round(ratio * 100))))

        # No matches found - not worth dealing with.
        if not title_ratios:
            return None

        title_ratios = sorted(title_ratios, key=lambda entry: entry[1], reverse=True)

        # Exact match.
        if title_ratios[0][1] == 1:
            return title_ratios[0][0]

        # Only one match found.
        if len(title_ratios) == 1:
            return title_ratios[0][0]

        # Multiple matches, title within a single match - return match.
        query_within = [t for t in title_ratios if query in t[0]]
        if len(query_within) == 1:
            return query_within[0][0]

        # Multiple matches.
        return title_ratios

    def find_all_episodes(self, title):
        """ Returns a tuple of lists, (all_viewed, all_episodes). """
        viewed_episodes, all_episodes = [], []  # season, episode, title.
        for show, data in self._recorded_shows.iteritems():
            dates_viewed, _ = data
            if dates_viewed and title in show:
                viewed_episodes.append(show)

            if title in show:
                all_episodes.append(show)
        return all_episodes, viewed_episodes

    def find_all_dates_viewed(self, title):
        """ Returns a sorted set (desc) of dates a show has been viewed. """
        viewed_dates = set()
        for show, data in self._recorded_shows.iteritems():
            if title not in show:  # Not an episode of a show we're searching for.
                continue

            dates_viewed, _ = data
            if not dates_viewed:
                continue

            viewed_dates |= set(dates_viewed)

        # Those damn Nones need to be pruned before sorting.
        return sorted([date for date in viewed_dates if date], reverse=True)

    def sort_episodes(self, episodes):
        """ Sorts episodes in descending order by season, episode.
            The `episodes` argument must be in the form: [(season, episode, title), ]

            @see find_all_episodes for retrieving episodes.
        """
        def sort_key(entry):
            season, episode, _ = entry
            if isinstance(episode, (int, float)):
                return (-int(season), int(-episode))

            # Episode is 1;2, we care about the 2nd value.
            return (-float(season), -float(episode.split(';')[-1]))

        return sorted(episodes, key=lambda entry: sort_key(entry))

    def viewed_up_to_date(self, title):
        """ True if last_viewed_episode == last_updated_episode. """
        recorded, viewed = self.find_all_episodes(title)
        recorded, viewed = map(self.sort_episodes, [recorded, viewed])
        if not viewed:
            return False

        return recorded[0] == viewed[0]

    @property
    def titles(self):
        return list(set(map(lambda record: record[-1], self._recorded_shows.keys())))

    @property
    def entries(self):
        return self._recorded_shows

    @property
    def load_success(self):
        return self._load_success

    @property
    def all_viewed_episodes(self):
        return [(k, v) for k, v in self._recorded_shows.iteritems() if v[0]]
